@@ -1,6 +1,6 @@
 /*
   spamdyke -- a filter for stopping spam at connection time.
-  Copyright (C) 2012 Sam Clippinger (samc (at) silence (dot) org)
+  Copyright (C) 2014 Sam Clippinger (samc (at) silence (dot) org)
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
@@ -558,8 +558,17 @@ int nihdns_create_packet(struct filter_settings *current_settings, int id, char
 
   return_value = -1;
 
-  if ((target_name != NULL) &&
-      ((strlen_target_name = strlen(target_name)) > 0))
+  if (target_name != NULL)
+    {
+    strlen_target_name = strlen(target_name);
+    while ((strlen_target_name > 0) &&
+           (target_name[strlen_target_name - 1] == '.'))
+      strlen_target_name--;
+    }
+  else
+    strlen_target_name = 0;
+
+  if (strlen_target_name > 0)
     {
     if (length_return_query > (sizeof(packet_template) + strlen_target_name + 10))
       {
@@ -622,7 +631,7 @@ int nihdns_create_packet(struct filter_settings *current_settings, int id, char
  *   NO RESULT (negative query): 0
  *   FOUND RESULT: length of answer
 */
-int nihdns_query(struct filter_settings *current_settings, char **target_name_array, int type_array, char *return_answer, int return_answer_length, char **return_answer_start, int *return_target_name_index)
+int nihdns_query(struct filter_settings *current_settings, char **target_name_array, int type_array, int preferred_type, char *return_answer, int return_answer_length, char **return_answer_start, int *return_target_name_index)
   {
   static unsigned short query_id = 0;
   static int config_type_array[] = CONFIG_DNS_TYPE_ARRAY;
@@ -670,6 +679,11 @@ int nihdns_query(struct filter_settings *current_settings, char **target_name_ar
   int num_types;
   int types[NUM_NIHDNS_TYPE];
   int current_timeout_total_secs;
+  char *potential_return_answer;
+  char *potential_return_answer_start;
+  int potential_return_target_name_index;
+  int potential_return_value;
+  int target_type;
 
   return_value = -1;
   error_occurred = 0;
@@ -682,6 +696,11 @@ int nihdns_query(struct filter_settings *current_settings, char **target_name_ar
   tcp_buf = NULL;
   tcp_buf_strlen = NULL;
   tcp_answer_len = NULL;
+  potential_return_answer = NULL;
+  potential_return_answer_start = NULL;
+  potential_return_target_name_index = -1;
+  potential_return_value = -1;
+  target_type = 0;
 
   if (query_id == 0)
     query_id = (int)random();
@@ -692,6 +711,12 @@ int nihdns_query(struct filter_settings *current_settings, char **target_name_ar
       {
       types[num_types] = nihdns_type_array[i];
       num_types++;
+
+      if (preferred_type == nihdns_type_array[i])
+        {
+        SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_DNS_PREFERRED_TYPE, nihdns_type_name(preferred_type));
+        target_type = preferred_type;
+        }
       }
 
   if ((num_types > 0) &&
@@ -724,6 +749,13 @@ int nihdns_query(struct filter_settings *current_settings, char **target_name_ar
                 {
                 for (i = 0; i < (num_names * num_types); i++)
                   tcp_answer_len[i] = 0;
+
+                if ((target_type != 0) &&
+                    ((potential_return_answer = (char *)malloc(sizeof(char) * return_answer_length)) == NULL))
+                  {
+                  SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_MALLOC, (unsigned long)(sizeof(char) * return_answer_length));
+                  error_occurred = 1;
+                  }
                 }
               else
                 {
@@ -1112,14 +1144,35 @@ int nihdns_query(struct filter_settings *current_settings, char **target_name_ar
                             SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_DNS_RECEIVED_TYPE, nihdns_type_name(type), nihdns_type_name(types[response_id % num_types]));
                             if (type == types[response_id % num_types])
                               {
-                              memcpy(return_answer, answer, MINVAL(response_length, return_answer_length));
-                              if (return_answer_start != NULL)
-                                *return_answer_start = return_answer + (answer_start - answer);
-                              if (return_target_name_index != NULL)
-                                *return_target_name_index = response_id / num_types;
+                              /*
+                               * If the caller asked for a specific "preferred" type and it hasn't been received,
+                               * hold this answer for now, just in case the "preferred" type arrives.  If it doesn't
+                               * arrive before the next packets are due to be sent, we'll just go with what we have.
+                               * Otherwise, any answer will do.
+                               */
+                              if ((target_type != 0) &&
+                                  (type != target_type))
+                                {
+                                SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_DNS_POTENTIAL_ANSWER, nihdns_type_name(type), nihdns_type_name(target_type));
 
-                              return_value = MINVAL(response_length, return_answer_length);
-                              break;
+                                memcpy(potential_return_answer, answer, MINVAL(response_length, return_answer_length));
+                                potential_return_answer_start = potential_return_answer + (answer_start - answer);
+                                potential_return_target_name_index = response_id / num_types;
+
+                                potential_return_value = MINVAL(response_length, return_answer_length);
+                                }
+                              else
+                                {
+                                memcpy(return_answer, answer, MINVAL(response_length, return_answer_length));
+                                if (return_answer_start != NULL)
+                                  *return_answer_start = return_answer + (answer_start - answer);
+                                if (return_target_name_index != NULL)
+                                  *return_target_name_index = response_id / num_types;
+
+                                return_value = MINVAL(response_length, return_answer_length);
+
+                                break;
+                                }
                               }
                             }
                         }
@@ -1183,6 +1236,18 @@ int nihdns_query(struct filter_settings *current_settings, char **target_name_ar
 
           if (return_value >= 0)
             break;
+          else if (potential_return_value >= 0)
+            {
+            memcpy(return_answer, potential_return_answer, potential_return_value);
+            if (return_answer_start != NULL)
+              *return_answer_start = return_answer + (potential_return_answer_start - potential_return_answer);
+            if (return_target_name_index != NULL)
+              *return_target_name_index = potential_return_target_name_index;
+
+            return_value = potential_return_value;
+
+            break;
+            }
           }
         else
           break;
@@ -1217,6 +1282,8 @@ int nihdns_query(struct filter_settings *current_settings, char **target_name_ar
     free(tcp_buf_strlen);
   if (tcp_answer_len != NULL)
     free(tcp_answer_len);
+  if (potential_return_answer != NULL)
+    free(potential_return_answer);
 
   return(return_value);
   }
@@ -1262,7 +1329,7 @@ int nihdns_rbl(struct filter_settings *current_settings, char **target_name_arra
 
   if ((target_name_array != NULL) &&
       (target_rbl_array != NULL) &&
-      (answer_length = nihdns_query(current_settings, target_name_array, current_settings->current_options->nihdns_query_type_rbl, answer, MAX_DNS_PACKET_BYTES_UDP, &answer_ptr, &target_name_index)) > 0)
+      (answer_length = nihdns_query(current_settings, target_name_array, current_settings->current_options->nihdns_query_type_rbl, 0, answer, MAX_DNS_PACKET_BYTES_UDP, &answer_ptr, &target_name_index)) > 0)
     {
     // This seems silly, but it defeats a strict-aliasing warning from gcc when
     // a more conventional cast is used: (struct nihdns_header *)&answer
@@ -1460,7 +1527,7 @@ int nihdns_ptr_lookup(struct filter_settings *current_settings, char *target_nam
   target_name_array[1] = NULL;
 
   if ((target_name != NULL) &&
-      ((answer_length = nihdns_query(current_settings, target_name_array, current_settings->current_options->nihdns_query_type_ptr, answer, MAX_DNS_PACKET_BYTES_UDP, &answer_ptr, NULL)) > 0))
+      ((answer_length = nihdns_query(current_settings, target_name_array, current_settings->current_options->nihdns_query_type_ptr, 0, answer, MAX_DNS_PACKET_BYTES_UDP, &answer_ptr, NULL)) > 0))
     {
     // This seems silly, but it defeats a strict-aliasing warning from gcc when
     // a more conventional cast is used: (struct nihdns_header *)&answer
@@ -1623,7 +1690,7 @@ int nihdns_a_inner(struct filter_settings *current_settings, char *target_name,
   target_name_array[1] = NULL;
 
   if ((target_name != NULL) &&
-      ((answer_length = nihdns_query(current_settings, target_name_array, target_query_type, answer, MAX_DNS_PACKET_BYTES_UDP, &answer_ptr, NULL)) > 0))
+      ((answer_length = nihdns_query(current_settings, target_name_array, target_query_type, 0, answer, MAX_DNS_PACKET_BYTES_UDP, &answer_ptr, NULL)) > 0))
     {
     // This seems silly, but it defeats a strict-aliasing warning from gcc when
     // a more conventional cast is used: (struct nihdns_header *)&answer
@@ -1777,7 +1844,7 @@ int nihdns_mx(struct filter_settings *current_settings, char *target_name, struc
   target_name_array[1] = NULL;
 
   if ((target_name != NULL) &&
-      ((answer_length = nihdns_query(current_settings, target_name_array, current_settings->current_options->nihdns_query_type_mx, answer, MAX_DNS_PACKET_BYTES_UDP, &answer_ptr, NULL)) > 0))
+      ((answer_length = nihdns_query(current_settings, target_name_array, current_settings->current_options->nihdns_query_type_mx, NIHDNS_TYPE_MX, answer, MAX_DNS_PACKET_BYTES_UDP, &answer_ptr, NULL)) > 0))
     {
     // This seems silly, but it defeats a strict-aliasing warning from gcc when
     // a more conventional cast is used: (struct nihdns_header *)&answer
@@ -1821,7 +1888,14 @@ int nihdns_mx(struct filter_settings *current_settings, char *target_name, struc
                   (ip_ints[3] <= 255))
                 {
                 SPAMDYKE_LOG_VERBOSE(current_settings, LOG_VERBOSE_MX_IP, host, target_name);
-                return_value = 1;
+
+                if ((ip_ints[0] != localhost_octets[0]) ||
+                    (ip_ints[1] != localhost_octets[1]) ||
+                    (ip_ints[2] != localhost_octets[2]) ||
+                    (ip_ints[3] != localhost_octets[3]))
+                  return_value = 1;
+                else
+                  SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_DNS_MX_LOCALHOST, host);
                 }
               else
                 return_value = nihdns_a_inner(current_settings, (char *)host, NULL, NULL, 1, current_settings->current_options->nihdns_query_type_a);
@@ -1867,12 +1941,12 @@ int nihdns_mx(struct filter_settings *current_settings, char *target_name, struc
             exit_loop = 1;
             break;
           }
-      }
-    else
-      {
-      SPAMDYKE_LOG_VERBOSE(current_settings, LOG_VERBOSE_DNS_COMPRESSION, target_name);
-      break;
-      }
+        }
+      else
+        {
+        SPAMDYKE_LOG_VERBOSE(current_settings, LOG_VERBOSE_DNS_COMPRESSION, target_name);
+        break;
+        }
 
     for (i = 0; (i < num_cnames) && !return_value; i++)
       if ((size = nihdns_expand(answer, answer + answer_length, cname_ptr[i], (char *)host, MAX_HOSTNAME)) >= 0)
