@@ -1,6 +1,6 @@
 /*
   spamdyke -- a filter for stopping spam at connection time.
-  Copyright (C) 2012 Sam Clippinger (samc (at) silence (dot) org)
+  Copyright (C) 2014 Sam Clippinger (samc (at) silence (dot) org)
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
@@ -439,7 +439,7 @@ int exec_checkpassword(struct filter_settings *current_settings, char *command_l
  *   ERROR: -1
  *   SUCCESS: length of returned string
  */
-int exec_command_argv(struct filter_settings *current_settings, char *filename, char *argv[], struct expect_send *protocol, char **return_content, int size_return_content)
+int exec_command_argv(struct filter_settings *current_settings, char *filename, char *argv[], struct expect_send *protocol, char **return_content, int size_return_content, int *return_status)
   {
   int return_value;
   int i;
@@ -525,10 +525,10 @@ int exec_command_argv(struct filter_settings *current_settings, char *filename,
 
                 if (((stdout_pipe[0] >= 0) &&
                      FD_ISSET(stdout_pipe[0], &read_fds) &&
-                     ((stdout_read_result = read(stdout_pipe[0], input_buf + strlen_input_buf, MAX_BUF - strlen_input_buf)) > 0)) ||
+                     ((stdout_read_result = read(stdout_pipe[0], input_buf + strlen_input_buf, MAX_COMMAND_BUF - strlen_input_buf)) > 0)) ||
                     ((stderr_pipe[0] >= 0) &&
                      FD_ISSET(stderr_pipe[0], &read_fds) &&
-                     ((stderr_read_result = read(stderr_pipe[0], input_buf + strlen_input_buf, MAX_BUF - strlen_input_buf)) > 0)))
+                     ((stderr_read_result = read(stderr_pipe[0], input_buf + strlen_input_buf, MAX_COMMAND_BUF - strlen_input_buf)) > 0)))
                   {
                   strlen_input_buf += stdout_read_result + stderr_read_result + 1;
                   strlen_input_total += stdout_read_result + stderr_read_result + 1;
@@ -655,8 +655,14 @@ int exec_command_argv(struct filter_settings *current_settings, char *filename,
               if (stderr_pipe[0] != -1)
                 close(stderr_pipe[0]);
 
-              if ((waitpid(child_pid, &status, WNOHANG) != child_pid) ||
-                  !WIFEXITED(status))
+              if ((waitpid(child_pid, &status, WNOHANG) == child_pid) &&
+                  WIFEXITED(status))
+                {
+                SPAMDYKE_LOG_DEBUG(current_settings, LOG_DEBUG_COMMAND_EXIT, WEXITSTATUS(status), new_filename);
+                if (return_status != NULL)
+                  *return_status = WEXITSTATUS(status);
+                }
+              else
                 {
                 kill(child_pid, SIGKILL);
                 SPAMDYKE_LOG_VERBOSE(current_settings, LOG_VERBOSE_COMMAND_ABEND "%s", new_filename);
@@ -793,7 +799,7 @@ int exec_command_argv(struct filter_settings *current_settings, char *filename,
  *   ERROR: -1
  *   SUCCESS: length of returned string
  */
-int exec_command(struct filter_settings *current_settings, char *command_line, struct expect_send *protocol, char **return_content, int size_return_content)
+int exec_command(struct filter_settings *current_settings, char *command_line, struct expect_send *protocol, char **return_content, int size_return_content, int *return_status)
   {
   int return_value;
   int i;
@@ -838,7 +844,7 @@ int exec_command(struct filter_settings *current_settings, char *command_line, s
 
         child_argv[j] = NULL;
 
-        return_value = exec_command_argv(current_settings, child_argv[0], child_argv, protocol, return_content, size_return_content);
+        return_value = exec_command_argv(current_settings, child_argv[0], child_argv, protocol, return_content, size_return_content, return_status);
 
         free(child_argv);
         }
