@@ -1,6 +1,6 @@
 /*
   spamdyke -- a filter for stopping spam at connection time.
-  Copyright (C) 2012 Sam Clippinger (samc (at) silence (dot) org)
+  Copyright (C) 2014 Sam Clippinger (samc (at) silence (dot) org)
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
@@ -441,9 +441,9 @@ int output_writeln(struct filter_settings *current_settings, int action, int tar
           last_char = data[data_length - 1];
           }
         }
-
-      fflush(NULL);
       }
+
+    fflush(NULL);
     }
 
   last_action = action;
@@ -503,7 +503,8 @@ void spamdyke_log(struct filter_settings *current_settings, int target_level, in
               }
 
             va_start(tmp_va, format);
-            vsyslog(syslog_level, format, tmp_va);
+            vsnprintf(tmp_data, MAX_BUF, format, tmp_va);
+            syslog(syslog_level, "%s", tmp_data);
             va_end(tmp_va);
             }
 
@@ -584,3 +585,55 @@ char *canonicalize_log_text(char *target_buf, int strlen_target_buf, char *input
 
   return(return_value);
   }
+
+char *escape_log_text(char *target_text, int strlen_target_text)
+  {
+  static char return_value[MAX_BUF + 1];
+  int i;
+  int strlen_return;
+
+  strlen_return = 0;
+
+  if (target_text != NULL)
+    for (i = 0; (i < strlen_target_text) && (strlen_return < (MAX_BUF - 1)); i++)
+      {
+      if (target_text[i] == '\0')
+        {
+        return_value[strlen_return] = '\\';
+        strlen_return++;
+        return_value[strlen_return] = '0';
+        }
+      else if (target_text[i] == '\r')
+        {
+        return_value[strlen_return] = '\\';
+        strlen_return++;
+        return_value[strlen_return] = 'r';
+        }
+      else if (target_text[i] == '\n')
+        {
+        return_value[strlen_return] = '\\';
+        strlen_return++;
+        return_value[strlen_return] = 'n';
+        }
+      else if (target_text[i] == '\t')
+        {
+        return_value[strlen_return] = '\\';
+        strlen_return++;
+        return_value[strlen_return] = 't';
+        }
+      else if (target_text[i] == '\\')
+        {
+        return_value[strlen_return] = '\\';
+        strlen_return++;
+        return_value[strlen_return] = '\\';
+        }
+      else
+        return_value[strlen_return] = target_text[i];
+
+      strlen_return++;
+      }
+
+  return_value[strlen_return] = '\0';
+
+  return(return_value);
+  }
