@@ -1,6 +1,6 @@
 /*
   spamdyke -- a filter for stopping spam at connection time.
-  Copyright (C) 2012 Sam Clippinger (samc (at) silence (dot) org)
+  Copyright (C) 2014 Sam Clippinger (samc (at) silence (dot) org)
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
@@ -25,6 +25,7 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <fnmatch.h>
+#include <unistd.h>
 #include "spamdyke.h"
 #include "environment.h"
 #include "log.h"
@@ -99,7 +100,7 @@ int find_path(struct filter_settings *current_settings, char *filename, char *en
       tmp_end = NULL;
       while (tmp_start != NULL)
         {
-        if ((tmp_end = strchr(tmp_start, ':')) != NULL)
+        if ((tmp_end = strchr(tmp_start, ENVIRONMENT_DELIMITER)) != NULL)
           {
           snprintf(new_filename, MAX_BUF, "%.*s" DIR_DELIMITER_STR "%s%n", (int)(tmp_end - tmp_start), tmp_start, filename, &strlen_filename);
           tmp_start = tmp_end + 1;
@@ -128,6 +129,59 @@ int find_path(struct filter_settings *current_settings, char *filename, char *en
   return(return_value);
   }
 
+/*
+ * When passed an entire command line string (e.g. from a .qmail file), this
+ * function finds just the command itself, allowing for escaping characters
+ * and quoting.
+ *
+ * RETURNS:
+ *   characters copied to return_text
+ */
+int find_command(char *input_text, char *return_text, int size_return_text)
+  {
+  int return_value;
+  int i;
+  int escape;
+  int inside_quote;
+
+  return_value = 0;
+  escape = 0;
+  inside_quote = 0;
+
+  if ((input_text != NULL) &&
+      (return_text != NULL))
+    {
+    for (i = 0; input_text[i] != '\0'; i++)
+      if (!isspace((int)input_text[i]))
+        break;
+
+    for (; (input_text[i] != '\0') && (return_value < size_return_text); i++)
+      if (isspace((int)input_text[i]) &&
+          !inside_quote &&
+          !escape)
+        break;
+      else if (escape)
+        {
+        return_text[return_value] = input_text[i];
+        return_value++;
+        escape = 0;
+        }
+      else if (input_text[i] == PATH_ESCAPE_CHAR)
+        escape = 1;
+      else if (input_text[i] == PATH_QUOTE_CHAR)
+        inside_quote = !inside_quote;
+      else
+        {
+        return_text[return_value] = input_text[i];
+        return_value++;
+        }
+
+    return_text[return_value] = '\0';
+    }
+
+  return(return_value);
+  }
+
 /*
  * Return value:
  *   NULL: no match
@@ -155,6 +209,18 @@ char *find_case_insensitive_needle(char *haystack, char *needle)
   }
 
 /*
+ * EXPECTS:
+ *   target_string = string content to search for, must not be NULL or zero length
+ *   strlen_target_string = length of target_string
+ *   target_entry = string content to search within, must not be NULL
+ *   strlen_target_entry = length of target_entry
+ *   start_wildcard = wildcard character that, if present as the first character in target_string, allows target_string to match in the middle of target_entry.  \0 disables this feature.
+ *   start_wildcard_matches = if start_wildcard matches the first character in target_string, a match is not found unless the character in target_entry immediately preceding the location is found
+ *     within start_wildcard_matches.  E.g. start_wildcard is '.', target_string is ".foo" and target_entry is "bar-foo-baz".  A match is found if start_wildcard_matches is "-!#" but not if it is
+ *     "/.,".  NULL disables this feature.
+ *   end_wildcard = like start_wildcard but at the end of target_string instead of the start
+ *   end_wildcard_matches = like start_wildcard_matches but at the end of target_string instead of the start
+ *
  * Return value:
  *   0: no match
  *   1: match found
@@ -220,6 +286,17 @@ int examine_entry(char *target_string, int strlen_target_string, char *target_en
   }
 
 /*
+ * EXPECTS:
+ *   search_filename = the file to search, line by line
+ *   target_string = string content to search for, must not be NULL or zero length
+ *   strlen_target_string = length of target_string
+ *   start_wildcard = wildcard character that, if present as the first character in target_string, allows target_string to match in the middle of a line.  \0 disables this feature.
+ *   start_wildcard_matches = if start_wildcard matches the first character in target_string, a match is not found unless the character in the line immediately preceding the location is found
+ *     within start_wildcard_matches.  E.g. start_wildcard is '.', target_string is ".foo" and the line is "bar-foo-baz".  A match is found if start_wildcard_matches is "-!#" but not if it is
+ *     "/.,".  NULL disables this feature.
+ *   end_wildcard = like start_wildcard but at the end of target_string instead of the start
+ *   end_wildcard_matches = like start_wildcard_matches but at the end of target_string instead of the start
+ *
  * Return value:
  *   ERROR: -1
  *   NOT FOUND: 0
@@ -510,7 +587,6 @@ int examine_tcprules_entry(struct filter_settings *current_settings, char *desti
  *   NOT FOUND: 0
  *   FOUND: matching line number
  */
-//FIXME: make this actually search the file for the best match possible according to the tcprules docs, not just the first match.
 int search_tcprules_file(struct filter_settings *current_settings, char *destination, int size_destination, char *search_filename, char *target_ip, char *target_name, int strlen_target_name)
   {
   int return_value;
@@ -836,7 +912,7 @@ char *search_domain_directory(struct filter_settings *current_settings, char *st
  *   ERROR: -1
  *   SUCCESS: number of lines read, excluding skipped lines (1-based)
  */
-int read_file(struct filter_settings *current_settings, char *target_filename, char ***return_content, int start_index, int start_line, int end_line)
+int read_file(struct filter_settings *current_settings, char *target_filename, char ***return_content, int start_index, int start_line, int end_line, int all_lines)
   {
   int return_value;
   int i;
@@ -873,8 +949,10 @@ int read_file(struct filter_settings *current_settings, char *target_filename, c
               (line_num < end_line)))
         {
         if ((fscanf(tmp_file, "%" STRINGIFY(MAX_FILE_BUF) "[^\r\n]", file_buf) == 1) &&
-            (file_buf[0] != COMMENT_DELIMITER) &&
-            ((strlen_line = strlen(file_buf)) > 0) &&
+            ((strlen_line = strlen(file_buf)) || 1) &&
+            (all_lines ||
+             ((file_buf[0] != COMMENT_DELIMITER) &&
+              (strlen_line > 0))) &&
             (line_num >= zero_start))
           {
           SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_READ_LINE, strlen_line, target_filename, line_num + 1, file_buf);
@@ -974,7 +1052,7 @@ int read_file_first_line(struct filter_settings *current_settings, char *target_
   tmp_ptr = tmp_array;
 
   if ((return_content != NULL) &&
-      (read_file(current_settings, target_filename, (char ***)&tmp_ptr, 0, 1, 1) != -1))
+      (read_file(current_settings, target_filename, (char ***)&tmp_ptr, 0, 1, 1, 0) != -1))
     {
     *return_content = tmp_array[0];
     return_value = (tmp_array[0] != NULL) ? strlen(tmp_array[0]) : 0;
@@ -1161,3 +1239,136 @@ int search_header_file(struct filter_settings *current_settings, char *search_fi
 
   return(return_value);
   }
+
+/*
+ * EXPECTS:
+ *   type_flag should be a value within the S_IFMT mask in stat.h -- S_IFDIR, S_IFREG, etc
+ *   permission_flags should be an OR of FILE_PERMISSION_READ, FILE_PERMISSION_WRITE or FILE_PERMISSION_EXECUTE, not bitshifted for user/group/other.
+ *   if target_uid is -1, geteuid() will be used
+ *   if target_gid is -1, getegid() and secondary groups will be used
+ *
+ * RETURNS:
+ *  -1: error
+ *  0: file does not exist, is the wrong type or does not have permissions
+ *  1: file exists, is the correct type and has the permissions
+ */
+int check_path_perms(struct filter_settings *current_settings, char *target_path, int type_flag, int permission_flags, struct stat *target_stat, int target_uid, int target_gid)
+  {
+  int return_value;
+  int i;
+  struct stat *tmp_stat;
+  struct stat internal_stat;
+  int found_match;
+  int num_groups;
+  gid_t *tmp_gid;
+
+  return_value = 0;
+  tmp_gid = NULL;
+
+  if (target_path != NULL)
+    {
+    tmp_stat = (target_stat != NULL) ? target_stat : &internal_stat;
+    if (!stat(target_path, tmp_stat))
+      {
+      SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_FILE_STAT, tmp_stat->st_mode, type_flag, tmp_stat->st_uid, tmp_stat->st_gid, target_path);
+
+      if (((type_flag == 0) ||
+           ((tmp_stat->st_mode & S_IFMT) == type_flag)) &&
+          (((permission_flags & FILE_PERMISSION_SETUID) == 0) ||
+           (tmp_stat->st_mode & S_ISUID)) &&
+          (((permission_flags & FILE_PERMISSION_SETGID) == 0) ||
+           (tmp_stat->st_mode & S_ISGID)) &&
+          (((permission_flags & FILE_PERMISSION_STICKY) == 0) ||
+           (tmp_stat->st_mode & S_ISVTX)))
+        {
+        if ((permission_flags & FILE_PERMISSION_READ) ||
+            (permission_flags & FILE_PERMISSION_WRITE) ||
+            (permission_flags & FILE_PERMISSION_EXECUTE))
+          {
+          if (((target_uid != -1) &&
+               (tmp_stat->st_uid == target_uid)) ||
+              ((target_uid == -1) &&
+               (tmp_stat->st_uid == geteuid())))
+            {
+            if ((((permission_flags & FILE_PERMISSION_READ) == 0) ||
+                 (tmp_stat->st_mode & S_IRUSR)) &&
+                (((permission_flags & FILE_PERMISSION_WRITE) == 0) ||
+                 (tmp_stat->st_mode & S_IWUSR)) &&
+                (((permission_flags & FILE_PERMISSION_EXECUTE) == 0) ||
+                 (tmp_stat->st_mode & S_IXUSR)))
+              return_value = 1;
+            }
+          else
+            {
+            found_match = 0;
+
+            if (target_gid != -1)
+              {
+              if (tmp_stat->st_gid == target_gid)
+                found_match = 1;
+              }
+            else if (tmp_stat->st_gid == getegid())
+              found_match = 1;
+            else if ((num_groups = getgroups(0, NULL)) > 0)
+              {
+              if ((tmp_gid = (gid_t *)malloc(sizeof(gid_t) * num_groups)) != NULL)
+                {
+                if (getgroups(num_groups, tmp_gid) > 0)
+                  for (i = 0; i < num_groups; i++)
+                    if (tmp_stat->st_gid == tmp_gid[i])
+                      {
+                      /* Secondary group owns the directory. */
+                      found_match = 1;
+                      break;
+                      }
+
+                free(tmp_gid);
+                tmp_gid = NULL;
+                }
+              else
+                {
+                SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_MALLOC, (long)(sizeof(gid_t) * num_groups));
+                return_value = -1;
+                }
+              }
+
+            if (found_match)
+              {
+              if ((((permission_flags & FILE_PERMISSION_READ) == 0) ||
+                   (tmp_stat->st_mode & S_IRGRP)) &&
+                  (((permission_flags & FILE_PERMISSION_WRITE) == 0) ||
+                   (tmp_stat->st_mode & S_IWGRP)) &&
+                  (((permission_flags & FILE_PERMISSION_EXECUTE) == 0) ||
+                   (tmp_stat->st_mode & S_IXGRP)))
+                return_value = 1;
+              }
+            else
+              if ((((permission_flags & FILE_PERMISSION_READ) == 0) ||
+                   (tmp_stat->st_mode & S_IROTH)) &&
+                  (((permission_flags & FILE_PERMISSION_WRITE) == 0) ||
+                   (tmp_stat->st_mode & S_IWOTH)) &&
+                  (((permission_flags & FILE_PERMISSION_EXECUTE) == 0) ||
+                   (tmp_stat->st_mode & S_IXOTH)))
+                return_value = 1;
+            }
+          }
+        else
+          return_value = 1;
+        }
+      }
+    else if ((errno == EACCES) ||
+             (errno == ENOENT) ||
+             (errno == ENOTDIR))
+      {
+      SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_FILE_STAT_FAIL, target_path, strerror(errno));
+      return_value = 0;
+      }
+    else
+      {
+      SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_STAT_ERRNO, target_path, strerror(errno));
+      return_value = -1;
+      }
+    }
+
+  return(return_value);
+  }
