@@ -1,6 +1,6 @@
 /*
   spamdyke -- a filter for stopping spam at connection time.
-  Copyright (C) 2012 Sam Clippinger (samc (at) silence (dot) org)
+  Copyright (C) 2014 Sam Clippinger (samc (at) silence (dot) org)
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
@@ -179,6 +179,8 @@ int tls_init_inner(struct filter_settings *current_settings, SSL_CTX **target_tl
   int tmp_rand;
   int error_occurred;
   int tls_return;
+  FILE *tmp_file;
+  DH *tmp_dh;
 
   return_value = 0;
 
@@ -203,7 +205,7 @@ int tls_init_inner(struct filter_settings *current_settings, SSL_CTX **target_tl
       if (i < MAX_RAND_SEED)
         initialized = 1;
       else
-        SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_INIT);
+        SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_INIT, NULL);
       }
     else
       initialized = 1;
@@ -259,20 +261,68 @@ int tls_init_inner(struct filter_settings *current_settings, SSL_CTX **target_tl
           }
         }
 
-      if (SSL_CTX_check_private_key(*target_tls_context))
-        SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_TLS_CERT_CHECK, NULL);
-      else
+      if (!error_occurred)
         {
-        SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_CERT_CHECK "%s : %s", (current_settings->current_options->tls_privatekey_file != NULL) ? current_settings->current_options->tls_privatekey_file : current_settings->current_options->tls_certificate_file, tls_error(current_settings, 0));
-        error_occurred = 1;
+        if (SSL_CTX_check_private_key(*target_tls_context))
+          SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_TLS_CERT_CHECK, NULL);
+        else
+          {
+          SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_CERT_CHECK "%s : %s", (current_settings->current_options->tls_privatekey_file != NULL) ? current_settings->current_options->tls_privatekey_file : current_settings->current_options->tls_certificate_file, tls_error(current_settings, 0));
+          error_occurred = 1;
+          }
         }
 
-      if ((tls_return = SSL_CTX_set_cipher_list(*target_tls_context, current_settings->current_options->tls_cipher_list)) == 1)
-        SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_TLS_CIPHER_LIST, current_settings->current_options->tls_cipher_list);
-      else
+      if (!error_occurred &&
+          (current_settings->current_options->tls_dhparams_file != NULL))
         {
-        SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_CIPHER_LIST "%s: %s", current_settings->current_options->tls_cipher_list, tls_error(current_settings, tls_return));
-        error_occurred = 1;
+        if ((SSL_CTX_set_options(*target_tls_context, SSL_OP_SINGLE_DH_USE) & SSL_OP_SINGLE_DH_USE) == SSL_OP_SINGLE_DH_USE)
+          {
+          SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_TLS_OPTIONS, "SSL_OP_SINGLE_DH_USE");
+
+          if ((tmp_file = fopen(current_settings->current_options->tls_dhparams_file, "r")) != NULL)
+            {
+            if ((tmp_dh = PEM_read_DHparams(tmp_file, NULL, NULL, NULL)) != NULL)
+              {
+              SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_TLS_DHPARAMS, current_settings->current_options->tls_dhparams_file);
+
+              if ((tls_return = SSL_CTX_set_tmp_dh(*target_tls_context, tmp_dh)) == 1)
+                SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_TLS_SET_DHPARAMS, NULL);
+              else
+                {
+                SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_SET_DHPARAMS ": %s", tls_error(current_settings, tls_return));
+                error_occurred = 1;
+                }
+              }
+            else
+              {
+              SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_DHPARAMS "%s", current_settings->current_options->tls_dhparams_file);
+              error_occurred = 1;
+              }
+
+            fclose(tmp_file);
+            }
+          else
+            {
+            SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_OPEN "%s: %s", current_settings->current_options->tls_dhparams_file, strerror(errno));
+            error_occurred = 1;
+            }
+          }
+        else
+          {
+          SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_OPTIONS "SSL_OP_SINGLE_DH_USE", NULL);
+          error_occurred = 1;
+          }
+        }
+
+      if (!error_occurred)
+        {
+        if ((tls_return = SSL_CTX_set_cipher_list(*target_tls_context, current_settings->current_options->tls_cipher_list)) == 1)
+          SPAMDYKE_LOG_EXCESSIVE(current_settings, LOG_DEBUGX_TLS_CIPHER_LIST, current_settings->current_options->tls_cipher_list);
+        else
+          {
+          SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_CIPHER_LIST "%s: %s", current_settings->current_options->tls_cipher_list, tls_error(current_settings, tls_return));
+          error_occurred = 1;
+          }
         }
 
       if (!error_occurred &&
@@ -280,14 +330,14 @@ int tls_init_inner(struct filter_settings *current_settings, SSL_CTX **target_tl
         return_value = 1;
       else
         {
-        SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_INIT);
+        SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_INIT, NULL);
 
         SSL_CTX_free(*target_tls_context);
         *target_tls_context = NULL;
         }
       }
     else
-      SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_INIT);
+      SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_INIT, NULL);
     }
 
   return(return_value);
@@ -364,6 +414,8 @@ int tls_start(struct filter_settings *current_settings, int read_fd, int write_f
         SSL_set_mode(current_settings->tls_session, SSL_MODE_AUTO_RETRY);
         current_settings->tls_state = TLS_STATE_ACTIVE_SPAMDYKE;
 
+        SPAMDYKE_LOG_DEBUG(current_settings, LOG_DEBUG_TLS_CIPHER, SSL_CIPHER_get_name(SSL_get_current_cipher(current_settings->tls_session)), SSL_CIPHER_get_bits(SSL_get_current_cipher(current_settings->tls_session), NULL));
+
         return_value = 1;
         break;
         }
@@ -388,7 +440,7 @@ int tls_start(struct filter_settings *current_settings, int read_fd, int write_f
             select(0, NULL, NULL, NULL, &tmp_timeval);
             break;
           default:
-            SPAMDYKE_LOG_VERBOSE(current_settings, LOG_ERROR_TLS_ACCEPT ": %s", tls_error(current_settings, tls_return));
+            SPAMDYKE_LOG_VERBOSE(current_settings, LOG_VERBOSE_TLS_ACCEPT ": %s", tls_error(current_settings, tls_return));
             continue_looping = 0;
 
             break;
@@ -396,10 +448,10 @@ int tls_start(struct filter_settings *current_settings, int read_fd, int write_f
 
     if (continue_looping &&
         (return_value == 0))
-      SPAMDYKE_LOG_VERBOSE(current_settings, LOG_ERROR_TLS_ACCEPT ": %s", tls_error(current_settings, tls_return));
+      SPAMDYKE_LOG_VERBOSE(current_settings, LOG_VERBOSE_TLS_ACCEPT ": %s", tls_error(current_settings, tls_return));
     }
   else
-    SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_INIT);
+    SPAMDYKE_LOG_ERROR(current_settings, LOG_ERROR_TLS_INIT, NULL);
 
   return(return_value);
   }
