@@ -1,6 +1,6 @@
 /*
   spamdyke -- a filter for stopping spam at connection time.
-  Copyright (C) 2012 Sam Clippinger (samc (at) silence (dot) org)
+  Copyright (C) 2014 Sam Clippinger (samc (at) silence (dot) org)
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
@@ -20,6 +20,6 @@
 
 #include "spamdyke.h"
 
-void usage(struct filter_settings *current_settings, int full_text, char *error_pattern, ...);
+void usage(struct filter_settings *current_settings, int text_level, char *error_pattern, ...);
 
 #endif /* USAGE_H */
